import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(scope="module")
def grains_filter_by_lookup(salt_ssh_cli):
    ret = salt_ssh_cli.run("grains.get", "os")
    assert ret.returncode == 0
    assert ret.data
    os = ret.data

    return {
        "common": {
            "has_common": True,
        },
        "custom_default": {
            "defaulted": True,
        },
        "merge": {
            "merged": True,
        },
        os: {
            "filtered": True,
        },
    }


@pytest.fixture(scope="module")
def grains_filter_by_default():
    return {
        "common": {
            "has_common": True,
        },
        "custom_default": {
            "defaulted": True,
        },
        "merge": {
            "merged": True,
        },
    }


@pytest.fixture(scope="module")
def grains_filter_by_states(
    salt_master, salt_ssh_cli, grains_filter_by_lookup, grains_filter_by_default
):
    filter_content = f"{{%- set lookup = {grains_filter_by_lookup} %}}"
    default_content = f"{{%- set lookup = {grains_filter_by_default} %}}"
    content = r"""
{%- set res = salt["grains.filter_by"](lookup, grain="os", merge=lookup["merge"], base="common", default="custom_default") %}
grains-filter-by:
  file.managed:
    - name: {{ salt["temp.file"]() }}
    - context: {{ res | json }}
    """
    try:
        with salt_master.state_tree.base.temp_file(
            "grains_filter_by.sls", filter_content + content
        ):
            with salt_master.state_tree.base.temp_file(
                "grains_filter_by_default.sls", default_content + content
            ):
                ret = salt_ssh_cli.run("--regen-thin", "test.true")
                assert ret.returncode == 0
                yield
    finally:
        salt_ssh_cli.run("--regen-thin", "test.true")


def test_grains_id(salt_ssh_cli):
    """
    Test salt-ssh grains id work for localhost.
    """
    ret = salt_ssh_cli.run("grains.get", "id")
    assert ret.returncode == 0
    assert ret.data == "localhost"


def test_grains_items(salt_ssh_cli):
    """
    test grains.items with salt-ssh
    """
    ret = salt_ssh_cli.run("grains.items")
    assert ret.returncode == 0
    assert ret.data
    assert isinstance(ret.data, dict)
    if salt.utils.platform.is_darwin():
        grain = "Darwin"
    elif salt.utils.platform.is_aix():
        grain = "AIX"
    elif salt.utils.platform.is_freebsd():
        grain = "FreeBSD"
    else:
        grain = "Linux"
    assert ret.data["kernel"] == grain


def test_grains_filter_by(salt_ssh_cli, grains_filter_by_lookup):
    """
    test grains.filter_by with salt-ssh
    """
    ret = salt_ssh_cli.run(
        "grains.filter_by",
        grains_filter_by_lookup,
        grain="os",
        merge=grains_filter_by_lookup["merge"],
        base="common",
        default="custom_default",
    )
    assert ret.returncode == 0
    assert ret.data
    assert "has_common" in ret.data
    assert "filtered" in ret.data
    assert "merged" in ret.data
    assert "defaulted" not in ret.data


@pytest.mark.usefixtures("grains_filter_by_states")
def test_grains_filter_by_jinja(salt_ssh_cli):
    """
    test grains.filter_by during template rendering with salt-ssh
    """
    ret = salt_ssh_cli.run("state.show_sls", "grains_filter_by")
    assert ret.returncode == 0
    assert ret.data
    rendered = ret.data["grains-filter-by"]["file"][1]["context"]

    assert "has_common" in rendered
    assert "filtered" in rendered
    assert "merged" in rendered
    assert "defaulted" not in rendered


def test_grains_filter_by_default(salt_ssh_cli, grains_filter_by_default):
    """
    test grains.filter_by with salt-ssh and default parameter
    """
    ret = salt_ssh_cli.run(
        "grains.filter_by",
        grains_filter_by_default,
        grain="os",
        merge=grains_filter_by_default["merge"],
        base="common",
        default="custom_default",
    )
    assert ret.returncode == 0
    assert ret.data
    assert "has_common" in ret.data
    assert "merged" in ret.data
    assert "defaulted" in ret.data


@pytest.mark.usefixtures("grains_filter_by_states")
def test_grains_filter_by_default_jinja(salt_ssh_cli, grains_filter_by_default):
    """
    test grains.filter_by during template rendering with salt-ssh and default parameter
    """
    ret = salt_ssh_cli.run("state.show_sls", "grains_filter_by_default")
    assert ret.returncode == 0
    assert ret.data
    rendered = ret.data["grains-filter-by"]["file"][1]["context"]

    assert "has_common" in rendered
    assert "merged" in rendered
    assert "defaulted" in rendered



import logging
import os
import shutil
import subprocess

import pytest

from tests.conftest import FIPS_TESTRUN

log = logging.getLogger(__name__)


@pytest.fixture(scope="package")
def salt_mm_master_1(request, salt_factories):

    subprocess.check_output(["ip", "addr", "add", "172.16.0.1/32", "dev", "lo"])

    config_defaults = {
        "open_mode": True,
        "transport": request.config.getoption("--transport"),
    }
    config_overrides = {
        "interface": "0.0.0.0",
        "master_sign_pubkey": True,
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        "mm-master-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    try:
        with factory.started(start_timeout=180):
            yield factory
    finally:

        try:
            subprocess.check_output(["ip", "addr", "del", "172.16.0.1/32", "dev", "lo"])
        except subprocess.CalledProcessError:
            pass


@pytest.fixture(scope="package")
def mm_master_1_salt_cli(salt_mm_master_1):
    return salt_mm_master_1.salt_cli(timeout=180)


@pytest.fixture(scope="package")
def salt_mm_master_2(salt_factories, salt_mm_master_1):
    # if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
    #    subprocess.check_output(["ifconfig", "lo0", "alias", "127.0.0.2", "up"])

    config_defaults = {
        "open_mode": True,
        "transport": salt_mm_master_1.config["transport"],
    }
    config_overrides = {
        "interface": "0.0.0.0",
        "master_sign_pubkey": True,
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }

    # Use the same ports for both masters, they are binding to different interfaces
    for key in (
        "ret_port",
        "publish_port",
    ):
        config_overrides[key] = salt_mm_master_1.config[key] + 1
    factory = salt_factories.salt_master_daemon(
        "mm-master-2",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )

    # Both masters will share the same signing key pair
    for keyfile in ("master_sign.pem", "master_sign.pub"):
        shutil.copyfile(
            os.path.join(salt_mm_master_1.config["pki_dir"], keyfile),
            os.path.join(factory.config["pki_dir"], keyfile),
        )
    with factory.started(start_timeout=180):
        yield factory


@pytest.fixture(scope="package")
def mm_master_2_salt_cli(salt_mm_master_2):
    return salt_mm_master_2.salt_cli(timeout=180)


@pytest.fixture(scope="package")
def salt_mm_minion_1(salt_mm_master_1, salt_mm_master_2, master_alive_interval):
    config_defaults = {
        "transport": salt_mm_master_1.config["transport"],
    }

    mm_master_1_port = salt_mm_master_1.config["ret_port"]
    mm_master_2_port = salt_mm_master_2.config["ret_port"]
    config_overrides = {
        "master": [
            f"master1.local:{mm_master_1_port}",
            f"master2.local:{mm_master_2_port}",
        ],
        "publish_port": salt_mm_master_1.config["publish_port"],
        "master_alive_interval": master_alive_interval,
        "master_tries": -1,
        "verify_master_pubkey_sign": True,
        "retry_dns": True,
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = salt_mm_master_1.salt_minion_daemon(
        "mm-minion-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    # Need to grab the public signing key from the master, either will do
    shutil.copyfile(
        os.path.join(salt_mm_master_1.config["pki_dir"], "master_sign.pub"),
        os.path.join(factory.config["pki_dir"], "master_sign.pub"),
    )
    # with factory.started(start_timeout=180):
    yield factory



"""
Execute salt convenience routines
"""

import logging
import os

import salt.exceptions
import salt.loader
import salt.minion
import salt.utils.args
import salt.utils.event
import salt.utils.files
import salt.utils.user
from salt.client import mixins
from salt.output import display_output
from salt.utils.lazy import verify_fun

log = logging.getLogger(__name__)


class RunnerClient(mixins.SyncClientMixin, mixins.AsyncClientMixin):
    """
    The interface used by the :command:`salt-run` CLI tool on the Salt Master

    It executes :ref:`runner modules <all-salt.runners>` which run on the Salt
    Master.

    Importing and using ``RunnerClient`` must be done on the same machine as
    the Salt Master and it must be done using the same user that the Salt
    Master is running as.

    Salt's :conf_master:`external_auth` can be used to authenticate calls. The
    eauth user must be authorized to execute runner modules: (``@runner``).
    Only the :py:meth:`master_call` below supports eauth.
    """

    client = "runner"
    tag_prefix = "run"

    @property
    def functions(self):
        if not hasattr(self, "_functions"):
            if not hasattr(self, "utils"):
                self.utils = salt.loader.utils(self.opts)
            # Must be self.functions for mixin to work correctly :-/
            try:
                self._functions = salt.loader.runner(
                    self.opts, utils=self.utils, context=self.context
                )
            except AttributeError:
                # Just in case self.utils is still not present (perhaps due to
                # problems with the loader), load the runner funcs without them
                self._functions = salt.loader.runner(self.opts, context=self.context)

        return self._functions

    def _reformat_low(self, low):
        """
        Format the low data for RunnerClient()'s master_call() function

        This also normalizes the following low data formats to a single, common
        low data structure.

        Old-style low: ``{'fun': 'jobs.lookup_jid', 'jid': '1234'}``
        New-style: ``{'fun': 'jobs.lookup_jid', 'kwarg': {'jid': '1234'}}``
        CLI-style: ``{'fun': 'jobs.lookup_jid', 'arg': ['jid="1234"']}``
        """
        fun = low.pop("fun")
        verify_fun(self.functions, fun)

        eauth_creds = {
            i: low.pop(i)
            for i in [
                "username",
                "password",
                "eauth",
                "token",
                "client",
                "user",
                "key",
            ]
            if i in low
        }

        # Run name=value args through parse_input. We don't need to run kwargs
        # through because there is no way to send name=value strings in the low
        # dict other than by including an `arg` array.
        _arg, _kwarg = salt.utils.args.parse_input(low.pop("arg", []), condition=False)
        _kwarg.update(low.pop("kwarg", {}))

        # If anything hasn't been pop()'ed out of low by this point it must be
        # an old-style kwarg.
        _kwarg.update(low)

        # Finally, mung our kwargs to a format suitable for the byzantine
        # load_args_and_kwargs so that we can introspect the function being
        # called and fish for invalid kwargs.
        munged = []
        munged.extend(_arg)
        munged.append(dict(__kwarg__=True, **_kwarg))
        arg, kwarg = salt.minion.load_args_and_kwargs(
            self.functions[fun], munged, ignore_invalid=True
        )

        return dict(fun=fun, kwarg={"kwarg": kwarg, "arg": arg}, **eauth_creds)

    def cmd_async(self, low):
        """
        Execute a runner function asynchronously; eauth is respected

        This function requires that :conf_master:`external_auth` is configured
        and the user is authorized to execute runner functions: (``@runner``).

        .. code-block:: python

            runner.cmd_async({
                'fun': 'jobs.list_jobs',
                'username': 'saltdev',
                'password': 'saltdev',
                'eauth': 'pam',
            })
        """
        reformatted_low = self._reformat_low(low)

        return mixins.AsyncClientMixin.cmd_async(self, reformatted_low)

    def cmd_sync(self, low, timeout=None, full_return=False):
        """
        Execute a runner function synchronously; eauth is respected

        This function requires that :conf_master:`external_auth` is configured
        and the user is authorized to execute runner functions: (``@runner``).

        .. code-block:: python

            runner.cmd_sync({
                'fun': 'jobs.list_jobs',
                'username': 'saltdev',
                'password': 'saltdev',
                'eauth': 'pam',
            })
        """
        reformatted_low = self._reformat_low(low)
        return mixins.SyncClientMixin.cmd_sync(
            self, reformatted_low, timeout, full_return
        )

    def cmd(
        self,
        fun,
        arg=None,
        pub_data=None,
        kwarg=None,
        print_event=True,
        full_return=False,
    ):  # pylint: disable=useless-super-delegation
        """
        Execute a function

        .. code-block:: python

            >>> opts = salt.config.master_config('/etc/salt/master')
            >>> runner = salt.runner.RunnerClient(opts)
            >>> runner.cmd('jobs.list_jobs', [])
            {
                '20131219215650131543': {
                    'Arguments': [300],
                    'Function': 'test.sleep',
                    'StartTime': '2013, Dec 19 21:56:50.131543',
                    'Target': '*',
                    'Target-type': 'glob',
                    'User': 'saltdev'
                },
                '20131219215921857715': {
                    'Arguments': [300],
                    'Function': 'test.sleep',
                    'StartTime': '2013, Dec 19 21:59:21.857715',
                    'Target': '*',
                    'Target-type': 'glob',
                    'User': 'saltdev'
                },
            }

        """
        return super().cmd(fun, arg, pub_data, kwarg, print_event, full_return)


class Runner(RunnerClient):
    """
    Execute the salt runner interface
    """

    def __init__(self, opts, context=None):
        super().__init__(opts, context=context)
        self.returners = salt.loader.returners(opts, self.functions, context=context)
        self.outputters = salt.loader.outputters(opts)

    def print_docs(self):
        """
        Print out the documentation!
        """
        arg = self.opts.get("fun", None)
        docs = super().get_docs(arg)
        for fun in sorted(docs):
            display_output(f"{fun}:", "text", self.opts)
            print(docs[fun])

    # TODO: move to mixin whenever we want a salt-wheel cli
    def run(self, full_return=False):
        """
        Execute the runner sequence
        """
        import salt.minion

        ret = {}
        if self.opts.get("doc", False):
            self.print_docs()
        else:
            low = {"fun": self.opts["fun"]}
            try:
                # Allocate a jid
                async_pub = self._gen_async_pub(jid=self.opts.get("jid"))
                self.jid = async_pub["jid"]

                fun_args = salt.utils.args.parse_input(
                    self.opts["arg"], no_parse=self.opts.get("no_parse", [])
                )

                verify_fun(self.functions, low["fun"])
                args, kwargs = salt.minion.load_args_and_kwargs(
                    self.functions[low["fun"]], fun_args
                )
                low["arg"] = args
                low["kwarg"] = kwargs

                if self.opts.get("eauth"):
                    if "token" in self.opts:
                        try:
                            with salt.utils.files.fopen(
                                os.path.join(self.opts["cachedir"], ".root_key"), "r"
                            ) as fp_:
                                low["key"] = salt.utils.stringutils.to_unicode(
                                    fp_.readline()
                                )
                        except OSError:
                            low["token"] = self.opts["token"]

                    # If using eauth and a token hasn't already been loaded into
                    # low, prompt the user to enter auth credentials
                    if "token" not in low and "key" not in low and self.opts["eauth"]:
                        # This is expensive. Don't do it unless we need to.
                        import salt.auth

                        resolver = salt.auth.Resolver(self.opts)
                        res = resolver.cli(self.opts["eauth"])
                        if self.opts["mktoken"] and res:
                            tok = resolver.token_cli(self.opts["eauth"], res)
                            if tok:
                                low["token"] = tok.get("token", "")
                        if not res:
                            log.error("Authentication failed")
                            return ret
                        low.update(res)
                        low["eauth"] = self.opts["eauth"]
                else:
                    user = salt.utils.user.get_specific_user()

                if low["fun"] in ["state.orchestrate", "state.orch", "state.sls"]:
                    low["kwarg"]["orchestration_jid"] = async_pub["jid"]

                # Run the runner!
                if self.opts.get("async", False):
                    if self.opts.get("eauth"):
                        async_pub = self.cmd_async(low)
                    else:
                        async_pub = self.asynchronous(
                            self.opts["fun"], low, user=user, pub=async_pub
                        )
                    # by default: info will be not enougth to be printed out !
                    log.warning(
                        "Running in asynchronous mode. Results of this execution may "
                        "be collected by attaching to the master event bus or "
                        "by examining the master job cache, if configured. "
                        "This execution is running under tag %s",
                        async_pub["tag"],
                    )
                    return async_pub["jid"]  # return the jid

                # otherwise run it in the main process
                if self.opts.get("show_jid"):
                    print(f"jid: {self.jid}")

                if self.opts.get("eauth"):
                    ret = self.cmd_sync(low)
                    if isinstance(ret, dict) and set(ret) == {"data", "outputter"}:
                        outputter = ret["outputter"]
                        ret = ret["data"]
                    else:
                        outputter = None
                    display_output(ret, outputter, self.opts)
                else:
                    ret = self._proc_function(
                        instance=self,
                        opts=self.opts,
                        fun=self.opts["fun"],
                        low=low,
                        user=user,
                        tag=async_pub["tag"],
                        jid=async_pub["jid"],
                        daemonize=False,
                        full_return=full_return,
                    )
            except salt.exceptions.SaltException as exc:
                with salt.utils.event.get_event("master", opts=self.opts) as evt:
                    evt.fire_event(
                        {
                            "success": False,
                            "return": f"{exc}",
                            "retcode": 254,
                            "fun": self.opts["fun"],
                            "fun_args": fun_args,
                            "jid": self.jid,
                        },
                        tag=f"salt/run/{self.jid}/ret",
                    )
                # Attempt to grab documentation
                if "fun" in low:
                    ret = self.get_docs("{}*".format(low["fun"]))
                else:
                    ret = None

                # If we didn't get docs returned then
                # return the `not availble` message.
                if not ret:
                    ret = f"{exc}"
                if not self.opts.get("quiet", False):
                    display_output(ret, "nested", self.opts)
            else:
                log.debug("Runner return: %s", ret)

            return ret



import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]


def test_list_all(client, salt_minion, salt_sub_minion):
    ret = client.cmd("key.list_all", print_event=False)
    assert ret
    assert "minions" in ret
    assert salt_minion.id in ret["minions"]
    assert salt_sub_minion.id in ret["minions"]


def test_gen(client):
    ret = client.cmd(
        "key.gen", kwarg={"id_": "soundtechniciansrock"}, print_event=False
    )
    assert ret
    assert "pub" in ret
    try:
        assert ret["pub"].startswith("-----BEGIN PUBLIC KEY-----")
    except AttributeError:
        assert ret["pub"].startswith("-----BEGIN RSA PUBLIC KEY-----")

    assert "priv" in ret
    assert ret["priv"].startswith("-----BEGIN RSA PRIVATE KEY-----")


def test_master_key_str(client):
    ret = client.cmd("key.master_key_str", print_event=False)
    assert ret
    assert "local" in ret
    data = ret["local"]
    assert "master.pub" in data
    assert data["master.pub"].startswith("-----BEGIN PUBLIC KEY-----")



"""
Service support for RHEL-based systems, including support for both upstart and sysvinit

.. important::
    If you feel that Salt should be using this module to manage services on a
    minion, and it is using a different module (or gives an error similar to
    *'service.start' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import fnmatch
import glob
import logging
import os
import re
import stat

import salt.utils.path

log = logging.getLogger(__name__)

__func_alias__ = {"reload_": "reload"}

# Define the module's virtual name
__virtualname__ = "service"

# Import upstart module if needed
HAS_UPSTART = False
if salt.utils.path.which("initctl"):
    try:
        # Don't re-invent the wheel, import the helper functions from the
        # upstart module.
        from salt.modules.upstart_service import (
            _upstart_disable,
            _upstart_enable,
            _upstart_is_enabled,
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Unable to import helper functions from salt.modules.upstart: %s", exc
        )
    else:
        HAS_UPSTART = True


def __virtual__():
    """
    Only work on select distros which still use Red Hat's /usr/bin/service for
    management of either sysvinit or a hybrid sysvinit/upstart init system.
    """
    # Disable when booted with systemd
    if __utils__["systemd.booted"](__context__):
        return (
            False,
            "The rh_service execution module failed to load: this system was booted"
            " with systemd.",
        )

    # Enable on these platforms only.
    enable = {
        "XenServer",
        "XCP-ng",
        "RedHat",
        "CentOS",
        "ScientificLinux",
        "CloudLinux",
        "Amazon",
        "Fedora",
        "ALT",
        "OEL",
        "SUSE  Enterprise Server",
        "SUSE",
        "McAfee  OS Server",
        "VirtuozzoLinux",
    }
    if __grains__["os"] in enable:

        if __grains__["os"] == "SUSE":
            if str(__grains__["osrelease"]).startswith("11"):
                return __virtualname__
            else:
                return (False, "Cannot load rh_service module on SUSE > 11")

        osrelease_major = __grains__.get("osrelease_info", [0])[0]

        if __grains__["os"] in ("XenServer", "XCP-ng"):
            if osrelease_major >= 7:
                return (
                    False,
                    "XenServer and XCP-ng >= 7 use systemd, will not load rh_service.py"
                    " as virtual 'service'",
                )
            return __virtualname__

        if __grains__["os"] == "Fedora":
            if osrelease_major >= 15:
                return (
                    False,
                    "Fedora >= 15 uses systemd, will not load rh_service.py "
                    "as virtual 'service'",
                )
        if __grains__["os"] in (
            "RedHat",
            "CentOS",
            "ScientificLinux",
            "OEL",
            "CloudLinux",
        ):
            if osrelease_major >= 7:
                return (
                    False,
                    "RedHat-based distros >= version 7 use systemd, will not "
                    "load rh_service.py as virtual 'service'",
                )
        return __virtualname__
    return (False, f"Cannot load rh_service module: OS not in {enable}")


def _runlevel():
    """
    Return the current runlevel
    """
    out = __salt__["cmd.run"]("/sbin/runlevel")
    # unknown will be returned while inside a kickstart environment, since
    # this is usually a server deployment it should be safe to assume runlevel
    # 3.  If not all service related states will throw an out of range
    # exception here which will cause other functions to fail.
    if "unknown" in out:
        return "3"
    else:
        return out.split()[1]


def _chkconfig_add(name):
    """
    Run 'chkconfig --add' for a service whose script is installed in
    /etc/init.d.  The service is initially configured to be disabled at all
    run-levels.
    """
    cmd = f"/sbin/chkconfig --add {name}"
    if __salt__["cmd.retcode"](cmd, python_shell=False) == 0:
        log.info('Added initscript "%s" to chkconfig', name)
        return True
    else:
        log.error('Unable to add initscript "%s" to chkconfig', name)
        return False


def _service_is_upstart(name):
    """
    Return True if the service is an upstart service, otherwise return False.
    """
    return HAS_UPSTART and os.path.exists(f"/etc/init/{name}.conf")


def _service_is_sysv(name):
    """
    Return True if the service is a System V service (includes those managed by
    chkconfig); otherwise return False.
    """
    try:
        # Look for user-execute bit in file mode.
        return bool(os.stat(os.path.join("/etc/init.d", name)).st_mode & stat.S_IXUSR)
    except OSError:
        return False


def _service_is_chkconfig(name):
    """
    Return True if the service is managed by chkconfig.
    """
    cmdline = f"/sbin/chkconfig --list {name}"
    return (
        __salt__["cmd.retcode"](cmdline, python_shell=False, ignore_retcode=True) == 0
    )


def _sysv_is_enabled(name, runlevel=None):
    """
    Return True if the sysv (or chkconfig) service is enabled for the specified
    runlevel; otherwise return False.  If `runlevel` is None, then use the
    current runlevel.
    """
    # Try chkconfig first.
    result = _chkconfig_is_enabled(name, runlevel)
    if result:
        return True

    if runlevel is None:
        runlevel = _runlevel()
    return len(glob.glob(f"/etc/rc.d/rc{runlevel}.d/S??{name}")) > 0


def _chkconfig_is_enabled(name, runlevel=None):
    """
    Return ``True`` if the service is enabled according to chkconfig; otherwise
    return ``False``.  If ``runlevel`` is ``None``, then use the current
    runlevel.
    """
    cmdline = f"/sbin/chkconfig --list {name}"
    result = __salt__["cmd.run_all"](cmdline, python_shell=False)

    if runlevel is None:
        runlevel = _runlevel()
    if result["retcode"] == 0:
        for row in result["stdout"].splitlines():
            if f"{runlevel}:on" in row:
                if row.split()[0] == name:
                    return True
            elif row.split() == [name, "on"]:
                return True
    return False


def _sysv_enable(name):
    """
    Enable the named sysv service to start at boot.  The service will be enabled
    using chkconfig with default run-levels if the service is chkconfig
    compatible.  If chkconfig is not available, then this will fail.
    """
    if not _service_is_chkconfig(name) and not _chkconfig_add(name):
        return False
    cmd = f"/sbin/chkconfig {name} on"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def _sysv_disable(name):
    """
    Disable the named sysv service from starting at boot.  The service will be
    disabled using chkconfig with default run-levels if the service is chkconfig
    compatible; otherwise, the service will be disabled for the current
    run-level only.
    """
    if not _service_is_chkconfig(name) and not _chkconfig_add(name):
        return False
    cmd = f"/sbin/chkconfig {name} off"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def _sysv_delete(name):
    """
    Delete the named sysv service from the system. The service will be
    deleted using chkconfig.
    """
    if not _service_is_chkconfig(name):
        return False
    cmd = f"/sbin/chkconfig --del {name}"
    return not __salt__["cmd.retcode"](cmd)


def _upstart_delete(name):
    """
    Delete an upstart service. This will only rename the .conf file
    """
    if HAS_UPSTART:
        if os.path.exists(f"/etc/init/{name}.conf"):
            os.rename(
                f"/etc/init/{name}.conf",
                f"/etc/init/{name}.conf.removed",
            )
    return True


def _upstart_services():
    """
    Return list of upstart services.
    """
    if HAS_UPSTART:
        return [os.path.basename(name)[:-5] for name in glob.glob("/etc/init/*.conf")]
    else:
        return []


def _sysv_services():
    """
    Return list of sysv services.
    """
    _services = []
    output = __salt__["cmd.run"](["chkconfig", "--list"], python_shell=False)
    for line in output.splitlines():
        comps = line.split()
        try:
            if comps[1].startswith("0:"):
                _services.append(comps[0])
        except IndexError:
            continue
    # Return only the services that have an initscript present
    return [x for x in _services if _service_is_sysv(x)]


def get_enabled(limit=""):
    """
    Return the enabled services. Use the ``limit`` param to restrict results
    to services of that type.

    CLI Examples:

    .. code-block:: bash

        salt '*' service.get_enabled
        salt '*' service.get_enabled limit=upstart
        salt '*' service.get_enabled limit=sysvinit
    """
    limit = limit.lower()
    if limit == "upstart":
        return sorted(name for name in _upstart_services() if _upstart_is_enabled(name))
    elif limit == "sysvinit":
        runlevel = _runlevel()
        return sorted(
            name for name in _sysv_services() if _sysv_is_enabled(name, runlevel)
        )
    else:
        runlevel = _runlevel()
        return sorted(
            [name for name in _upstart_services() if _upstart_is_enabled(name)]
            + [name for name in _sysv_services() if _sysv_is_enabled(name, runlevel)]
        )


def get_disabled(limit=""):
    """
    Return the disabled services. Use the ``limit`` param to restrict results
    to services of that type.

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_disabled
        salt '*' service.get_disabled limit=upstart
        salt '*' service.get_disabled limit=sysvinit
    """
    limit = limit.lower()
    if limit == "upstart":
        return sorted(
            name for name in _upstart_services() if not _upstart_is_enabled(name)
        )
    elif limit == "sysvinit":
        runlevel = _runlevel()
        return sorted(
            name for name in _sysv_services() if not _sysv_is_enabled(name, runlevel)
        )
    else:
        runlevel = _runlevel()
        return sorted(
            [name for name in _upstart_services() if not _upstart_is_enabled(name)]
            + [
                name
                for name in _sysv_services()
                if not _sysv_is_enabled(name, runlevel)
            ]
        )


def get_all(limit=""):
    """
    Return all installed services. Use the ``limit`` param to restrict results
    to services of that type.

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
        salt '*' service.get_all limit=upstart
        salt '*' service.get_all limit=sysvinit
    """
    limit = limit.lower()
    if limit == "upstart":
        return sorted(_upstart_services())
    elif limit == "sysvinit":
        return sorted(_sysv_services())
    else:
        return sorted(_sysv_services() + _upstart_services())


def available(name, limit=""):
    """
    Return True if the named service is available.  Use the ``limit`` param to
    restrict results to services of that type.

    CLI Examples:

    .. code-block:: bash

        salt '*' service.available sshd
        salt '*' service.available sshd limit=upstart
        salt '*' service.available sshd limit=sysvinit
    """
    if limit == "upstart":
        return _service_is_upstart(name)
    elif limit == "sysvinit":
        return _service_is_sysv(name)
    else:
        return (
            _service_is_upstart(name)
            or _service_is_sysv(name)
            or _service_is_chkconfig(name)
        )


def missing(name, limit=""):
    """
    The inverse of service.available.
    Return True if the named service is not available.  Use the ``limit`` param to
    restrict results to services of that type.

    CLI Examples:

    .. code-block:: bash

        salt '*' service.missing sshd
        salt '*' service.missing sshd limit=upstart
        salt '*' service.missing sshd limit=sysvinit
    """
    if limit == "upstart":
        return not _service_is_upstart(name)
    elif limit == "sysvinit":
        return not _service_is_sysv(name)
    else:
        if _service_is_upstart(name) or _service_is_sysv(name):
            return False
        else:
            return True


def start(name):
    """
    Start the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """
    if _service_is_upstart(name):
        cmd = f"start {name}"
    else:
        cmd = f"/sbin/service {name} start"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def stop(name):
    """
    Stop the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    if _service_is_upstart(name):
        cmd = f"stop {name}"
    else:
        cmd = f"/sbin/service {name} stop"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def restart(name):
    """
    Restart the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """
    if _service_is_upstart(name):
        cmd = f"restart {name}"
    else:
        cmd = f"/sbin/service {name} restart"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def reload_(name):
    """
    Reload the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.reload <service name>
    """
    if _service_is_upstart(name):
        cmd = f"reload {name}"
    else:
        cmd = f"/sbin/service {name} reload"
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def status(name, sig=None):
    """
    Return the status for a service.
    If the name contains globbing, a dict mapping service name to True/False
    values is returned.

    .. versionchanged:: 2018.3.0
        The service name can now be a glob (e.g. ``salt*``)

    Args:
        name (str): The name of the service to check
        sig (str): Signature to use to find the service via ps

    Returns:
        bool: True if running, False otherwise
        dict: Maps service name to True if running, False otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name> [service signature]
    """
    if sig:
        return bool(__salt__["status.pid"](sig))

    contains_globbing = bool(re.search(r"\*|\?|\[.+\]", name))
    if contains_globbing:
        services = fnmatch.filter(get_all(), name)
    else:
        services = [name]
    results = {}
    for service in services:
        if _service_is_upstart(service):
            cmd = f"status {service}"
            results[service] = "start/running" in __salt__["cmd.run"](
                cmd, python_shell=False
            )
        else:
            cmd = f"/sbin/service {service} status"
            results[service] = (
                __salt__["cmd.retcode"](cmd, python_shell=False, ignore_retcode=True)
                == 0
            )
    if contains_globbing:
        return results
    return results[name]


def delete(name, **kwargs):
    """
    Delete the named service

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' service.delete <service name>
    """
    if _service_is_upstart(name):
        return _upstart_delete(name)
    else:
        return _sysv_delete(name)


def enable(name, **kwargs):
    """
    Enable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable <service name>
    """
    if _service_is_upstart(name):
        return _upstart_enable(name)
    else:
        return _sysv_enable(name)


def disable(name, **kwargs):
    """
    Disable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable <service name>
    """
    if _service_is_upstart(name):
        return _upstart_disable(name)
    else:
        return _sysv_disable(name)


def enabled(name, **kwargs):
    """
    Check to see if the named service is enabled to start on boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled <service name>
    """
    if _service_is_upstart(name):
        return _upstart_is_enabled(name)
    else:
        return _sysv_is_enabled(name)


def disabled(name):
    """
    Check to see if the named service is disabled to start on boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled <service name>
    """
    if _service_is_upstart(name):
        return not _upstart_is_enabled(name)
    else:
        return not _sysv_is_enabled(name)



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def lgpo(modules):
    return modules.lgpo


def test_hierarchical_return(lgpo):
    result = lgpo.get_policy(
        policy_name="Calculator",
        policy_class="Machine",
        hierarchical_return=True,
    )
    result = result["Administrative Templates"]
    result = result["Windows Components"]
    result = result["Microsoft User Experience Virtualization"]
    result = result["Applications"]
    result = result["Calculator"]
    assert result in ("Enabled", "Disabled", "Not Configured")


def test_return_value_only_false(lgpo):
    result = lgpo.get_policy(
        policy_name="Calculator",
        policy_class="Machine",
        return_value_only=False,
    )
    assert result[
        r"Windows Components\Microsoft User Experience Virtualization\Applications\Calculator"
    ] in ("Enabled", "Disabled", "Not Configured")


def test_return_full_policy_names_false(lgpo):
    result = lgpo.get_policy(
        policy_name="Calculator",
        policy_class="Machine",
        return_full_policy_names=False,
        return_value_only=False,
    )
    assert result["Calculator"] in ("Enabled", "Disabled", "Not Configured")


def test_61860_calculator(lgpo):
    result = lgpo.get_policy(policy_name="Calculator", policy_class="Machine")
    # Any of the following are valid settings. We're only making sure it doesn't
    # throw a stacktrace
    assert result in ("Enabled", "Disabled", "Not Configured")



"""
Unit Tests for the mac_desktop execution module.
"""

import pytest

import salt.modules.mac_desktop as mac_desktop
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {mac_desktop: {}}


def test_get_output_volume():
    """
    Test if it get the output volume (range 0 to 100)
    """
    mock = MagicMock(return_value={"retcode": 0, "stdout": "25"})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        assert mac_desktop.get_output_volume() == "25"


def test_get_output_volume_error():
    """
    Tests that an error is raised when cmd.run_all errors
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        pytest.raises(CommandExecutionError, mac_desktop.get_output_volume)


def test_set_output_volume():
    """
    Test if it set the volume of sound (range 0 to 100)
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}), patch(
        "salt.modules.mac_desktop.get_output_volume", MagicMock(return_value="25")
    ):
        assert mac_desktop.set_output_volume("25")


def test_set_output_volume_error():
    """
    Tests that an error is raised when cmd.run_all errors
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        pytest.raises(CommandExecutionError, mac_desktop.set_output_volume, "25")


def test_screensaver():
    """
    Test if it launch the screensaver
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        assert mac_desktop.screensaver()


def test_screensaver_error():
    """
    Tests that an error is raised when cmd.run_all errors
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        pytest.raises(CommandExecutionError, mac_desktop.screensaver)


def test_lock():
    """
    Test if it lock the desktop session
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        assert mac_desktop.lock()


def test_lock_error():
    """
    Tests that an error is raised when cmd.run_all errors
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        pytest.raises(CommandExecutionError, mac_desktop.lock)


def test_say():
    """
    Test if it says some words.
    """
    mock = MagicMock(return_value={"retcode": 0})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        assert mac_desktop.say()


def test_say_error():
    """
    Tests that an error is raised when cmd.run_all errors
    """
    mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
        pytest.raises(CommandExecutionError, mac_desktop.say)



"""
Approximate the Unix find(1) command and return a list of paths that
meet the specified criteria.

The options include match criteria:
    name    = file-glob                 # case sensitive
    iname   = file-glob                 # case insensitive
    regex   = file-regex                # case sensitive
    iregex  = file-regex                # case insensitive
    type    = file-types                # match any listed type
    user    = users                     # match any listed user
    group   = groups                    # match any listed group
    size    = [+-]number[size-unit]     # default unit = byte
    mtime   = interval                  # modified since date
    grep    = regex                     # search file contents
and/or actions:
    delete [= file-types]               # default type = 'f'
    exec    = command [arg ...]         # where {} is replaced by pathname
    print  [= print-opts]
and/or depth criteria:
   maxdepth = maximum depth to transverse in path
   mindepth = minimum depth to transverse before checking files or directories

The default action is 'print=path'.

file-glob:
    *                = match zero or more chars
    ?                = match zero or single char
    [abc]            = match a, b, or c
    [!abc] or [^abc] = match anything except a, b, and c
    [x-y]            = match chars x through y
    [!x-y] or [^x-y] = match anything except chars x through y
    {a,b,c}          = match a or b or c

file-regex:
    a Python re (regular expression) pattern

file-types: a string of one or more of the following:
    a: all file types
    b: block device
    c: character device
    d: directory
    p: FIFO (named pipe)
    f: plain file
    l: symlink
    s: socket

users:
    a space and/or comma separated list of user names and/or uids

groups:
    a space and/or comma separated list of group names and/or gids

size-unit:
    b: bytes
    k: kilobytes
    m: megabytes
    g: gigabytes
    t: terabytes

interval:
    [<num>w] [<num>[d]] [<num>h] [<num>m] [<num>s]

    where:
        w: week
        d: day
        h: hour
        m: minute
        s: second

print-opts: a comma and/or space separated list of one or more of
the following:

    group: group name
    md5:   MD5 digest of file contents
    mode:  file permissions (as as integer)
    mtime: last modification time (as time_t)
    name:  file basename
    path:  file absolute path
    size:  file size in bytes
    type:  file type
    user:  user name
"""

import logging
import os
import re
import shutil
import stat
import sys
import time
from subprocess import PIPE, Popen

import salt.defaults.exitcodes
import salt.utils.args
import salt.utils.hashutils
import salt.utils.path
import salt.utils.stringutils
from salt.utils.filebuffer import BufferedReader

try:
    import grp
    import pwd

    # TODO: grp and pwd are both used in the code, we better make sure that
    # that code never gets run if importing them does not succeed
except ImportError:
    pass


# Set up logger
log = logging.getLogger(__name__)

_REQUIRES_PATH = 1
_REQUIRES_STAT = 2
_REQUIRES_CONTENTS = 4

_FILE_TYPES = {
    "b": stat.S_IFBLK,
    "c": stat.S_IFCHR,
    "d": stat.S_IFDIR,
    "f": stat.S_IFREG,
    "l": stat.S_IFLNK,
    "p": stat.S_IFIFO,
    "s": stat.S_IFSOCK,
    stat.S_IFBLK: "b",
    stat.S_IFCHR: "c",
    stat.S_IFDIR: "d",
    stat.S_IFREG: "f",
    stat.S_IFLNK: "l",
    stat.S_IFIFO: "p",
    stat.S_IFSOCK: "s",
}

_INTERVAL_REGEX = re.compile(
    r"""
                             ^\s*
                             (?P<modifier>[+-]?)
                             (?: (?P<week>   \d+ (?:\.\d*)? ) \s* [wW]  )? \s*
                             (?: (?P<day>    \d+ (?:\.\d*)? ) \s* [dD]  )? \s*
                             (?: (?P<hour>   \d+ (?:\.\d*)? ) \s* [hH]  )? \s*
                             (?: (?P<minute> \d+ (?:\.\d*)? ) \s* [mM]  )? \s*
                             (?: (?P<second> \d+ (?:\.\d*)? ) \s* [sS]  )? \s*
                             $
                             """,
    flags=re.VERBOSE,
)

_PATH_DEPTH_IGNORED = (os.path.sep, os.path.curdir, os.path.pardir)


def _parse_interval(value):
    """
    Convert an interval string like 1w3d6h into the number of seconds, time
    resolution (1 unit of the smallest specified time unit) and the modifier(
    '+', '-', or '').
        w = week
        d = day
        h = hour
        m = minute
        s = second
    """
    match = _INTERVAL_REGEX.match(str(value))
    if match is None:
        raise ValueError(f"invalid time interval: '{value}'")

    result = 0
    resolution = None
    for name, multiplier in [
        ("second", 1),
        ("minute", 60),
        ("hour", 60 * 60),
        ("day", 60 * 60 * 24),
        ("week", 60 * 60 * 24 * 7),
    ]:
        if match.group(name) is not None:
            result += float(match.group(name)) * multiplier
            if resolution is None:
                resolution = multiplier

    return result, resolution, match.group("modifier")


def _parse_size(value):
    scalar = value.strip()

    if scalar.startswith(("-", "+")):
        style = scalar[0]
        scalar = scalar[1:]
    else:
        style = "="

    if scalar:
        multiplier = {
            "b": 2**0,
            "k": 2**10,
            "m": 2**20,
            "g": 2**30,
            "t": 2**40,
        }.get(scalar[-1].lower())
        if multiplier:
            scalar = scalar[:-1].strip()
        else:
            multiplier = 1
    else:
        multiplier = 1

    try:
        num = int(scalar) * multiplier
    except ValueError:
        try:
            num = int(float(scalar) * multiplier)
        except ValueError:
            raise ValueError(f'invalid size: "{value}"')

    if style == "-":
        min_size = 0
        max_size = num
    elif style == "+":
        min_size = num
        max_size = sys.maxsize
    else:
        min_size = num
        max_size = num + multiplier - 1

    return min_size, max_size


class Option:
    """
    Abstract base class for all find options.
    """

    def requires(self):
        return _REQUIRES_PATH


class NameOption(Option):
    """
    Match files with a case-sensitive glob filename pattern.
    Note: this is the 'basename' portion of a pathname.
    The option name is 'name', e.g. {'name' : '*.txt'}.
    """

    def __init__(self, key, value):
        self.regex = re.compile(
            value.replace(".", "\\.").replace("?", ".?").replace("*", ".*") + "$"
        )

    def match(self, dirname, filename, fstat):
        return self.regex.match(filename)


class InameOption(Option):
    """
    Match files with a case-insensitive glob filename pattern.
    Note: this is the 'basename' portion of a pathname.
    The option name is 'iname', e.g. {'iname' : '*.TXT'}.
    """

    def __init__(self, key, value):
        self.regex = re.compile(
            value.replace(".", "\\.").replace("?", ".?").replace("*", ".*") + "$",
            re.IGNORECASE,
        )

    def match(self, dirname, filename, fstat):
        return self.regex.match(filename)


class RegexOption(Option):
    """
    Match files with a case-sensitive regular expression.
    Note: this is the 'basename' portion of a pathname.
    The option name is 'regex', e.g. {'regex' : '.*\\.txt'}.
    """

    def __init__(self, key, value):
        try:
            self.regex = re.compile(value)
        except re.error:
            raise ValueError(f'invalid regular expression: "{value}"')

    def match(self, dirname, filename, fstat):
        return self.regex.match(filename)


class IregexOption(Option):
    """
    Match files with a case-insensitive regular expression.
    Note: this is the 'basename' portion of a pathname.
    The option name is 'iregex', e.g. {'iregex' : '.*\\.txt'}.
    """

    def __init__(self, key, value):
        try:
            self.regex = re.compile(value, re.IGNORECASE)
        except re.error:
            raise ValueError(f'invalid regular expression: "{value}"')

    def match(self, dirname, filename, fstat):
        return self.regex.match(filename)


class TypeOption(Option):
    """
    Match files by their file type(s).
    The file type(s) are specified as an optionally comma and/or space
    separated list of letters.
        b = block device
        c = character device
        d = directory
        f = regular (plain) file
        l = symbolic link
        p = FIFO (named pipe)
        s = socket
    The option name is 'type', e.g. {'type' : 'd'} or {'type' : 'bc'}.
    """

    def __init__(self, key, value):
        # remove whitespace and commas
        value = "".join(value.strip().replace(",", "").split())
        self.ftypes = set()
        for ftype in value:
            try:
                self.ftypes.add(_FILE_TYPES[ftype])
            except KeyError:
                raise ValueError(f'invalid file type "{ftype}"')

    def requires(self):
        return _REQUIRES_STAT

    def match(self, dirname, filename, fstat):
        return stat.S_IFMT(fstat[stat.ST_MODE]) in self.ftypes


class OwnerOption(Option):
    """
    Match files by their owner name(s) and/or uid(s), e.g. 'root'.
    The names are a space and/or comma separated list of names and/or integers.
    A match occurs when the file's uid matches any user specified.
    The option name is 'owner', e.g. {'owner' : 'root'}.
    """

    def __init__(self, key, value):
        self.uids = set()
        for name in value.replace(",", " ").split():
            if name.isdigit():
                self.uids.add(int(name))
            else:
                try:
                    self.uids.add(pwd.getpwnam(value).pw_uid)
                except KeyError:
                    raise ValueError(f'no such user "{name}"')

    def requires(self):
        return _REQUIRES_STAT

    def match(self, dirname, filename, fstat):
        return fstat[stat.ST_UID] in self.uids


class GroupOption(Option):
    """
    Match files by their group name(s) and/or uid(s), e.g. 'admin'.
    The names are a space and/or comma separated list of names and/or integers.
    A match occurs when the file's gid matches any group specified.
    The option name is 'group', e.g. {'group' : 'admin'}.
    """

    def __init__(self, key, value):
        self.gids = set()
        for name in value.replace(",", " ").split():
            if name.isdigit():
                self.gids.add(int(name))
            else:
                try:
                    self.gids.add(grp.getgrnam(name).gr_gid)
                except KeyError:
                    raise ValueError(f'no such group "{name}"')

    def requires(self):
        return _REQUIRES_STAT

    def match(self, dirname, filename, fstat):
        return fstat[stat.ST_GID] in self.gids


class SizeOption(Option):
    """
    Match files by their size.
    Prefix the size with '-' to find files the specified size and smaller.
    Prefix the size with '+' to find files the specified size and larger.
    Without the +/- prefix, match the exact file size.
    The size can be suffixed with (case-insensitive) suffixes:
        b = bytes
        k = kilobytes
        m = megabytes
        g = gigabytes
        t = terabytes
    The option name is 'size', e.g. {'size' : '+1G'}.
    """

    def __init__(self, key, value):
        self.min_size, self.max_size = _parse_size(value)

    def requires(self):
        return _REQUIRES_STAT

    def match(self, dirname, filename, fstat):
        return self.min_size <= fstat[stat.ST_SIZE] <= self.max_size


class MtimeOption(Option):
    """
    Match files modified since the specified time.
    The option name is 'mtime', e.g. {'mtime' : '3d'}.
    The value format is [<num>w] [<num>[d]] [<num>h] [<num>m] [<num>s]
    where num is an integer or float and the case-insensitive suffixes are:
        w = week
        d = day
        h = hour
        m = minute
        s = second
    Whitespace is ignored in the value.
    """

    def __init__(self, key, value):
        secs, resolution, modifier = _parse_interval(value)
        self.mtime = time.time() - int(secs / resolution) * resolution
        self.modifier = modifier

    def requires(self):
        return _REQUIRES_STAT

    def match(self, dirname, filename, fstat):
        if self.modifier == "-":
            return fstat[stat.ST_MTIME] >= self.mtime
        else:
            return fstat[stat.ST_MTIME] <= self.mtime


class GrepOption(Option):
    """Match files when a pattern occurs within the file.
    The option name is 'grep', e.g. {'grep' : '(foo)|(bar}'}.
    """

    def __init__(self, key, value):
        try:
            self.regex = re.compile(value)
        except re.error:
            raise ValueError(f'invalid regular expression: "{value}"')

    def requires(self):
        return _REQUIRES_CONTENTS | _REQUIRES_STAT

    def match(self, dirname, filename, fstat):
        if not stat.S_ISREG(fstat[stat.ST_MODE]):
            return None
        dfilename = os.path.join(dirname, filename)
        with BufferedReader(dfilename, mode="rb") as bread:
            for chunk in bread:
                if self.regex.search(chunk):
                    return dfilename
        return None


class PrintOption(Option):
    """
    Return information about a matched file.
    Print options are specified as a comma and/or space separated list of
    one or more of the following:
        group  = group name
        md5    = MD5 digest of file contents
        mode   = file mode (as integer)
        mtime  = last modification time (as time_t)
        name   = file basename
        path   = file absolute path
        size   = file size in bytes
        type   = file type
        user   = user name
    """

    def __init__(self, key, value):
        self.need_stat = False
        self.print_title = False
        self.fmt = []
        for arg in value.replace(",", " ").split():
            self.fmt.append(arg)
            if arg not in ["name", "path"]:
                self.need_stat = True
        if not self.fmt:
            self.fmt.append("path")

    def requires(self):
        return _REQUIRES_STAT if self.need_stat else _REQUIRES_PATH

    def execute(self, fullpath, fstat, test=False):
        result = []
        for arg in self.fmt:
            if arg == "path":
                result.append(fullpath)
            elif arg == "name":
                result.append(os.path.basename(fullpath))
            elif arg == "size":
                result.append(fstat[stat.ST_SIZE])
            elif arg == "type":
                result.append(_FILE_TYPES.get(stat.S_IFMT(fstat[stat.ST_MODE]), "?"))
            elif arg == "mode":
                # PY3 compatibility: Use radix value 8 on int type-cast explicitly
                result.append(int(oct(fstat[stat.ST_MODE])[-3:], 8))
            elif arg == "mtime":
                result.append(fstat[stat.ST_MTIME])
            elif arg == "user":
                uid = fstat[stat.ST_UID]
                try:
                    result.append(pwd.getpwuid(uid).pw_name)
                except KeyError:
                    result.append(uid)
            elif arg == "group":
                gid = fstat[stat.ST_GID]
                try:
                    result.append(grp.getgrgid(gid).gr_name)
                except KeyError:
                    result.append(gid)
            elif arg == "md5":
                if stat.S_ISREG(fstat[stat.ST_MODE]):
                    md5digest = salt.utils.hashutils.get_hash(fullpath, "md5")
                    result.append(md5digest)
                else:
                    result.append("")

        if len(result) == 1:
            return result[0]
        else:
            return result


class DeleteOption(TypeOption):
    """
    Deletes matched file.
    Delete options are one or more of the following:
        a: all file types
        b: block device
        c: character device
        d: directory
        p: FIFO (named pipe)
        f: plain file
        l: symlink
        s: socket
    """

    def __init__(self, key, value):
        if "a" in value:
            value = "bcdpfls"
        super().__init__(key, value)

    def execute(self, fullpath, fstat, test=False):
        if test:
            return fullpath
        try:
            if os.path.isfile(fullpath) or os.path.islink(fullpath):
                os.remove(fullpath)
            elif os.path.isdir(fullpath):
                shutil.rmtree(fullpath)
        except OSError as exc:
            return None
        return fullpath


class ExecOption(Option):
    """
    Execute the given command, {} replaced by filename.
    Quote the {} if commands might include whitespace.
    """

    def __init__(self, key, value):
        self.command = value

    def execute(self, fullpath, fstat, test=False):
        try:
            command = self.command.replace("{}", fullpath)
            print(salt.utils.args.shlex_split(command))
            p = Popen(salt.utils.args.shlex_split(command), stdout=PIPE, stderr=PIPE)
            (out, err) = p.communicate()
            if err:
                log.error(
                    "Error running command: %s\n\n%s",
                    command,
                    salt.utils.stringutils.to_str(err),
                )
            return f"{command}:\n{salt.utils.stringutils.to_str(out)}\n"

        except Exception as e:  # pylint: disable=broad-except
            log.error('Exception while executing command "%s":\n\n%s', command, e)
            return f"{fullpath}: Failed"


class Finder:
    def __init__(self, options):
        self.actions = []
        self.maxdepth = None
        self.mindepth = 0
        self.test = False
        criteria = {
            _REQUIRES_PATH: list(),
            _REQUIRES_STAT: list(),
            _REQUIRES_CONTENTS: list(),
        }
        if "mindepth" in options:
            self.mindepth = options["mindepth"]
            del options["mindepth"]
        if "maxdepth" in options:
            self.maxdepth = options["maxdepth"]
            del options["maxdepth"]
        if "test" in options:
            self.test = options["test"]
            del options["test"]
        for key, value in options.items():
            if key.startswith("_"):
                # this is a passthrough object, continue
                continue
            if not value:
                raise ValueError(f'missing value for "{key}" option')
            try:
                obj = globals()[key.title() + "Option"](key, value)
            except KeyError:
                raise ValueError(f'invalid option "{key}"')
            if hasattr(obj, "match"):
                requires = obj.requires()
                if requires & _REQUIRES_CONTENTS:
                    criteria[_REQUIRES_CONTENTS].append(obj)
                elif requires & _REQUIRES_STAT:
                    criteria[_REQUIRES_STAT].append(obj)
                else:
                    criteria[_REQUIRES_PATH].append(obj)
            if hasattr(obj, "execute"):
                self.actions.append(obj)
        if not self.actions:
            self.actions.append(PrintOption("print", ""))
        # order criteria so that least expensive checks are done first
        self.criteria = (
            criteria[_REQUIRES_PATH]
            + criteria[_REQUIRES_STAT]
            + criteria[_REQUIRES_CONTENTS]
        )

    def find(self, path):
        """
        Generate filenames in path that satisfy criteria specified in
        the constructor.
        This method is a generator and should be repeatedly called
        until there are no more results.
        """
        if self.mindepth < 1:
            dirpath, name = os.path.split(path)
            match, fstat = self._check_criteria(dirpath, name, path)
            if match:
                yield from self._perform_actions(path, fstat=fstat)

        for dirpath, dirs, files in salt.utils.path.os_walk(path):
            relpath = os.path.relpath(dirpath, path)
            depth = path_depth(relpath) + 1
            if depth >= self.mindepth and (
                self.maxdepth is None or self.maxdepth >= depth
            ):
                for name in dirs + files:
                    fullpath = os.path.join(dirpath, name)
                    match, fstat = self._check_criteria(dirpath, name, fullpath)
                    if match:
                        yield from self._perform_actions(fullpath, fstat=fstat)

            if self.maxdepth is not None and depth > self.maxdepth:
                dirs[:] = []

    def _check_criteria(self, dirpath, name, fullpath, fstat=None):
        match = True
        for criterion in self.criteria:
            if fstat is None and criterion.requires() & _REQUIRES_STAT:
                try:
                    fstat = os.stat(fullpath)
                except OSError:
                    fstat = os.lstat(fullpath)
            if not criterion.match(dirpath, name, fstat):
                match = False
                break
        return match, fstat

    def _perform_actions(self, fullpath, fstat=None):
        for action in self.actions:
            if fstat is None and action.requires() & _REQUIRES_STAT:
                try:
                    fstat = os.stat(fullpath)
                except OSError:
                    fstat = os.lstat(fullpath)
            result = action.execute(fullpath, fstat, test=self.test)
            if result is not None:
                yield result


def path_depth(path):
    depth = 0
    head = path
    while True:
        head, tail = os.path.split(head)
        if not tail and (not head or head in _PATH_DEPTH_IGNORED):
            break
        if tail and tail not in _PATH_DEPTH_IGNORED:
            depth += 1
    return depth


def find(path, options):
    """
    WRITEME
    """
    finder = Finder(options)
    for path in finder.find(path):
        yield path


def _main():
    if len(sys.argv) < 2:
        sys.stderr.write(f"usage: {sys.argv[0]} path [options]\n")
        sys.exit(salt.defaults.exitcodes.EX_USAGE)

    path = sys.argv[1]
    criteria = {}

    for arg in sys.argv[2:]:
        key, value = arg.split("=")
        criteria[key] = value
    try:
        finder = Finder(criteria)
    except ValueError as ex:
        sys.stderr.write(f"error: {ex}\n")
        sys.exit(salt.defaults.exitcodes.EX_GENERIC)

    for result in finder.find(path):
        print(result)


if __name__ == "__main__":
    _main()



import pytest

import tests.support.netapi as netapi
from salt.netapi.rest_tornado import saltnado


@pytest.fixture
def client_config(client_config, netapi_port):
    client_config["rest_tornado"] = {"port": netapi_port}
    client_config["netapi_enable_clients"] = [
        "local",
        "local_async",
        "runner",
        "runner_async",
    ]
    return client_config


@pytest.fixture
def app(app_urls, load_auth, client_config, minion_config, salt_sub_minion):
    app = netapi.build_tornado_app(
        app_urls, load_auth, client_config, minion_config, setup_event_listener=True
    )
    try:
        yield app
    finally:
        app.event_listener.destroy()


@pytest.fixture
def client_headers(auth_token, content_type_map):
    return {
        "Content-Type": content_type_map["json"],
        saltnado.AUTH_TOKEN_HEADER: auth_token["token"],
    }


@pytest.fixture
def http_server(io_loop, app, client_headers, netapi_port):
    with netapi.TestsTornadoHttpServer(
        io_loop=io_loop, app=app, port=netapi_port, client_headers=client_headers
    ) as server:
        yield server


@pytest.fixture
def http_client(http_server):
    return http_server.client



import pytest

from salt.scripts import _pip_args, _pip_environment


def test_pip_environment_no_pypath():
    """
    We add PYTHONPATH to environemnt when it doesn't already exist.
    """
    extras = "/tmp/footest"
    env = {"HOME": "/home/dwoz"}
    pipenv = _pip_environment(env, extras)
    assert "PYTHONPATH" not in env
    assert "PYTHONPATH" in pipenv
    assert pipenv["PYTHONPATH"] == "/tmp/footest"


@pytest.mark.skip_on_windows(reason="Specific to *nix systems")
def test_pip_environment_pypath_nix():
    """
    We update PYTHONPATH in environemnt when it's already set.
    """
    extras = "/tmp/footest"
    env = {
        "HOME": "/home/dwoz",
        "PYTHONPATH": "/usr/local/lib/python3.10/site-packages",
    }
    assert "PYTHONPATH" in env
    pipenv = _pip_environment(env, extras)
    assert env["PYTHONPATH"] == "/usr/local/lib/python3.10/site-packages"
    assert "PYTHONPATH" in pipenv
    assert (
        pipenv["PYTHONPATH"] == "/tmp/footest:/usr/local/lib/python3.10/site-packages"
    )


@pytest.mark.skip_unless_on_windows(reason="Specific to win32 systems")
def test_pip_environment_pypath_win():
    """
    We update PYTHONPATH in environemnt when it's already set.
    """
    extras = "/tmp/footest"
    env = {
        "HOME": "/home/dwoz",
        "PYTHONPATH": "/usr/local/lib/python3.10/site-packages",
    }
    assert "PYTHONPATH" in env
    pipenv = _pip_environment(env, extras)
    assert env["PYTHONPATH"] == "/usr/local/lib/python3.10/site-packages"
    assert "PYTHONPATH" in pipenv
    assert (
        pipenv["PYTHONPATH"] == "/tmp/footest;/usr/local/lib/python3.10/site-packages"
    )


def test_pip_args_not_installing():
    extras = "/tmp/footest"
    args = ["list"]
    pargs = _pip_args(args, extras)
    assert pargs is not args
    assert args == ["list"]
    assert pargs == ["list"]


def test_pip_args_installing_without_target():
    extras = "/tmp/footest"
    args = ["install"]
    pargs = _pip_args(args, extras)
    assert pargs is not args
    assert args == ["install"]
    assert pargs == ["install", "--target=/tmp/footest"]


def test_pip_args_installing_with_target():
    extras = "/tmp/footest"
    args = ["install", "--target=/tmp/bartest"]
    pargs = _pip_args(args, extras)
    assert pargs is not args
    assert args == ["install", "--target=/tmp/bartest"]
    assert pargs == ["install", "--target=/tmp/bartest"]



import hashlib

import pytest
import tornado.iostream
from pytestshellutils.utils import ports

import salt.config
import salt.transport
import salt.transport.ipc
import salt.utils.asynchronous
import salt.utils.platform

pytestmark = [
    pytest.mark.core_test,
]


@pytest.fixture
def sock_dir(tmp_path):
    sock_dir_path = tmp_path / "test-socks"
    sock_dir_path.mkdir(parents=True, exist_ok=True)
    yield sock_dir_path


@pytest.fixture
def minion_config(sock_dir, minion_opts):
    minion_opts["sock_dir"] = sock_dir
    yield minion_opts


def test_ipc_connect_in_async_methods():
    "The connect method is in IPCMessageSubscriber's async_methods property"
    assert "connect" in salt.transport.ipc.IPCMessageSubscriber.async_methods


async def test_ipc_connect_sync_wrapped(io_loop, tmp_path):
    """
    Ensure IPCMessageSubscriber.connect gets wrapped by
    salt.utils.asynchronous.SyncWrapper.
    """
    if salt.utils.platform.is_windows():
        socket_path = ports.get_unused_localhost_port()
    else:
        socket_path = str(tmp_path / "noexist.ipc")
    subscriber = salt.utils.asynchronous.SyncWrapper(
        salt.transport.ipc.IPCMessageSubscriber,
        args=(socket_path,),
        kwargs={"io_loop": io_loop},
        loop_kwarg="io_loop",
    )
    with pytest.raises(tornado.iostream.StreamClosedError):
        # Don't `await subscriber.connect()`, that's the purpose of the SyncWrapper
        subscriber.connect()


@pytest.fixture
def master_config(sock_dir, master_opts):
    master_opts["sock_dir"] = sock_dir
    yield master_opts


@pytest.mark.skip_on_windows(reason="Unix socket not available on win32")
def test_master_ipc_server_unix(master_config, sock_dir):
    assert master_config.get("ipc_mode") != "tcp"
    server = salt.transport.ipc_publish_server("master", master_config)
    assert server.pub_path == str(sock_dir / "master_event_pub.ipc")
    assert server.pull_path == str(sock_dir / "master_event_pull.ipc")


@pytest.mark.skip_on_windows(reason="Unix socket not available on win32")
def test_minion_ipc_server_unix(minion_config, sock_dir):
    minion_config["id"] = "foo"
    id_hash = hashlib.sha256(
        salt.utils.stringutils.to_bytes(minion_config["id"])
    ).hexdigest()[:10]
    assert minion_config.get("ipc_mode") != "tcp"
    server = salt.transport.ipc_publish_server("minion", minion_config)
    assert server.pub_path == str(sock_dir / f"minion_event_{id_hash}_pub.ipc")
    assert server.pull_path == str(sock_dir / f"minion_event_{id_hash}_pull.ipc")


def test_master_ipc_server_tcp(master_config, sock_dir):
    master_config["ipc_mode"] = "tcp"
    server = salt.transport.ipc_publish_server("master", master_config)
    assert server.pub_host == "127.0.0.1"
    assert server.pub_port == int(master_config["tcp_master_pub_port"])
    assert server.pub_path is None
    assert server.pull_host == "127.0.0.1"
    assert server.pull_port == int(master_config["tcp_master_pull_port"])
    assert server.pull_path is None


def test_minion_ipc_server_tcp(minion_config, sock_dir):
    minion_config["ipc_mode"] = "tcp"
    server = salt.transport.ipc_publish_server("minion", minion_config)
    assert server.pub_host == "127.0.0.1"
    assert server.pub_port == int(minion_config["tcp_pub_port"])
    assert server.pub_path is None
    assert server.pull_host == "127.0.0.1"
    assert server.pull_port == int(minion_config["tcp_pull_port"])
    assert server.pull_path is None



"""
GNOME implementations
"""

import logging
import re

import salt.utils.path

try:
    import pwd

    HAS_PWD = True
except ImportError:
    HAS_PWD = False


try:
    from gi.repository import Gio, GLib  # pylint: disable=W0611

    HAS_GLIB = True
except ImportError:
    HAS_GLIB = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "gnome"

# Don't shadow built-in's.
__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Only load if the Gio and Glib modules are available
    """
    if HAS_PWD and HAS_GLIB:
        return __virtualname__
    return (
        False,
        "The gnome_desktop execution module cannot be loaded: "
        "The Gio and GLib modules are not available",
    )


class _GSettings:
    def __init__(self, user, schema, key):
        self.SCHEMA = schema
        self.KEY = key
        self.USER = user
        self.UID = None
        self.HOME = None

    @property
    def gsetting_command(self):
        """
        return the command to run the gsettings binary
        """
        if salt.utils.path.which_bin(["dbus-run-session"]):
            cmd = ["dbus-run-session", "--", "gsettings"]
        else:
            cmd = ["dbus-launch", "--exit-with-session", "gsettings"]
        return cmd

    def _get(self):
        """
        get the value for user in gsettings

        """
        user = self.USER
        try:
            uid = pwd.getpwnam(user).pw_uid
        except KeyError:
            log.info("User does not exist")
            return False

        cmd = self.gsetting_command + ["get", str(self.SCHEMA), str(self.KEY)]
        environ = {}
        environ["XDG_RUNTIME_DIR"] = f"/run/user/{uid}"
        result = __salt__["cmd.run_all"](
            cmd, runas=user, env=environ, python_shell=False
        )

        if "stdout" in result:
            if "uint32" in result["stdout"]:
                return re.sub("uint32 ", "", result["stdout"])
            else:
                return result["stdout"]
        else:
            return False

    def _set(self, value):
        """
        set the value for user in gsettings

        """
        user = self.USER
        try:
            uid = pwd.getpwnam(user).pw_uid
        except KeyError:
            log.info("User does not exist")
            result = {}
            result["retcode"] = 1
            result["stdout"] = f"User {user} does not exist"
            return result

        cmd = self.gsetting_command + ["set", self.SCHEMA, self.KEY, value]
        environ = {}
        environ["XDG_RUNTIME_DIR"] = f"/run/user/{uid}"
        result = __salt__["cmd.run_all"](
            cmd, runas=user, env=environ, python_shell=False
        )
        return result


def ping(**kwargs):
    """
    A test to ensure the GNOME module is loaded

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.ping user=<username>

    """
    return True


def getIdleDelay(**kwargs):
    """
    Return the current idle delay setting in seconds

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.getIdleDelay user=<username>

    """
    _gsession = _GSettings(
        user=kwargs.get("user"), schema="org.gnome.desktop.session", key="idle-delay"
    )
    return _gsession._get()


def setIdleDelay(delaySeconds, **kwargs):
    """
    Set the current idle delay setting in seconds

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.setIdleDelay <seconds> user=<username>

    """
    _gsession = _GSettings(
        user=kwargs.get("user"), schema="org.gnome.desktop.session", key="idle-delay"
    )
    return _gsession._set(delaySeconds)


def getClockFormat(**kwargs):
    """
    Return the current clock format, either 12h or 24h format.

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.getClockFormat user=<username>

    """
    _gsession = _GSettings(
        user=kwargs.get("user"),
        schema="org.gnome.desktop.interface",
        key="clock-format",
    )
    return _gsession._get()


def setClockFormat(clockFormat, **kwargs):
    """
    Set the clock format, either 12h or 24h format.

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.setClockFormat <12h|24h> user=<username>

    """
    if clockFormat != "12h" and clockFormat != "24h":
        return False
    _gsession = _GSettings(
        user=kwargs.get("user"),
        schema="org.gnome.desktop.interface",
        key="clock-format",
    )
    return _gsession._set(clockFormat)


def getClockShowDate(**kwargs):
    """
    Return the current setting, if the date is shown in the clock

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.getClockShowDate user=<username>

    """
    _gsession = _GSettings(
        user=kwargs.get("user"),
        schema="org.gnome.desktop.interface",
        key="clock-show-date",
    )
    return _gsession._get()


def setClockShowDate(kvalue, **kwargs):
    """
    Set whether the date is visible in the clock

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.setClockShowDate <True|False> user=<username>

    """
    if kvalue is not True and kvalue is not False:
        return False
    _gsession = _GSettings(
        user=kwargs.get("user"),
        schema="org.gnome.desktop.interface",
        key="clock-show-date",
    )
    return _gsession._set(kvalue)


def getIdleActivation(**kwargs):
    """
    Get whether the idle activation is enabled

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.getIdleActivation user=<username>

    """
    _gsession = _GSettings(
        user=kwargs.get("user"),
        schema="org.gnome.desktop.screensaver",
        key="idle-activation-enabled",
    )
    return _gsession._get()


def setIdleActivation(kvalue, **kwargs):
    """
    Set whether the idle activation is enabled

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.setIdleActivation <True|False> user=<username>

    """
    if kvalue is not True and kvalue is not False:
        return False
    _gsession = _GSettings(
        user=kwargs.get("user"),
        schema="org.gnome.desktop.screensaver",
        key="idle-activation-enabled",
    )
    return _gsession._set(kvalue)


def get(schema=None, key=None, user=None, **kwargs):
    """
    Get key in a particular GNOME schema

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.get user=<username> schema=org.gnome.desktop.screensaver key=idle-activation-enabled

    """
    _gsession = _GSettings(user=user, schema=schema, key=key)
    return _gsession._get()


def set_(schema=None, key=None, user=None, value=None, **kwargs):
    """
    Set key in a particular GNOME schema

    CLI Example:

    .. code-block:: bash

        salt '*' gnome.set user=<username> schema=org.gnome.desktop.screensaver key=idle-activation-enabled value=False

    """
    _gsession = _GSettings(user=user, schema=schema, key=key)
    return _gsession._set(value)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.modjk as modjk
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {modjk: {}}


def test_version():
    """
    Test for return the modjk version
    """
    with patch.object(
        modjk, "_do_http", return_value={"worker.jk_version": "mod_jk/1.2.37"}
    ):
        assert modjk.version() == "1.2.37"


def test_get_running():
    """
    Test for get the current running config (not from disk)
    """
    with patch.object(modjk, "_do_http", return_value={}):
        assert modjk.get_running() == {}


def test_dump_config():
    """
    Test for dump the original configuration that was loaded from disk
    """
    with patch.object(modjk, "_do_http", return_value={}):
        assert modjk.dump_config() == {}


def test_list_configured_members():
    """
    Test for return a list of member workers from the configuration files
    """
    with patch.object(modjk, "_do_http", return_value={}):
        assert modjk.list_configured_members("loadbalancer1") == []

    with patch.object(
        modjk,
        "_do_http",
        return_value={"worker.loadbalancer1.balance_workers": "SALT"},
    ):
        assert modjk.list_configured_members("loadbalancer1") == ["SALT"]


def test_workers():
    """
    Test for return a list of member workers and their status
    """
    with patch.object(modjk, "_do_http", return_value={"worker.list": "Salt1,Salt2"}):
        assert modjk.workers() == {}


def test_recover_all():
    """
    Test for set the all the workers in lbn to recover and
    activate them if they are not
    """
    with patch.object(modjk, "_do_http", return_value={}):
        assert modjk.recover_all("loadbalancer1") == {}

    with patch.object(
        modjk,
        "_do_http",
        return_value={"worker.loadbalancer1.balance_workers": "SALT"},
    ):
        with patch.object(
            modjk,
            "worker_status",
            return_value={"activation": "ACT", "state": "OK"},
        ):
            assert modjk.recover_all("loadbalancer1") == {
                "SALT": {"activation": "ACT", "state": "OK"}
            }


def test_reset_stats():
    """
    Test for reset all runtime statistics for the load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.reset_stats("loadbalancer1")


def test_lb_edit():
    """
    Test for edit the loadbalancer settings
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.lb_edit("loadbalancer1", {"vlr": 1, "vlt": 60})


def test_bulk_stop():
    """
    Test for stop all the given workers in the specific load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.bulk_stop(["node1", "node2", "node3"], "loadbalancer1")


def test_bulk_activate():
    """
    Test for activate all the given workers in the specific load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.bulk_activate(["node1", "node2", "node3"], "loadbalancer1")


def test_bulk_disable():
    """
    Test for disable all the given workers in the specific load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.bulk_disable(["node1", "node2", "node3"], "loadbalancer1")


def test_bulk_recover():
    """
    Test for recover all the given workers in the specific load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.bulk_recover(["node1", "node2", "node3"], "loadbalancer1")


def test_worker_status():
    """
    Test for return the state of the worker
    """
    with patch.object(
        modjk,
        "_do_http",
        return_value={"worker.node1.activation": "ACT", "worker.node1.state": "OK"},
    ):
        assert modjk.worker_status("node1") == {"activation": "ACT", "state": "OK"}

    with patch.object(modjk, "_do_http", return_value={}):
        assert not modjk.worker_status("node1")


def test_worker_recover():
    """
    Test for set the worker to recover this module will fail
    if it is in OK state
    """
    with patch.object(modjk, "_do_http", return_value={}):
        assert modjk.worker_recover("node1", "loadbalancer1") == {}


def test_worker_disable():
    """
    Test for set the worker to disable state in the lbn load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.worker_disable("node1", "loadbalancer1")


def test_worker_activate():
    """
    Test for set the worker to activate state in the lbn load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.worker_activate("node1", "loadbalancer1")


def test_worker_stop():
    """
    Test for set the worker to stopped state in the lbn load balancer
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.worker_stop("node1", "loadbalancer1")


def test_worker_edit():
    """
    Test for edit the worker settings
    """
    with patch.object(modjk, "_do_http", return_value={"worker.result.type": "OK"}):
        assert modjk.worker_edit("node1", "loadbalancer1", {"vwf": 500, "vwd": 60})



"""
the locale utils used by salt
"""

import sys

from salt.utils.decorators import memoize as real_memoize


@real_memoize
def get_encodings():
    """
    return a list of string encodings to try
    """
    encodings = [__salt_system_encoding__]

    try:
        sys_enc = sys.getdefaultencoding()
    except ValueError:  # system encoding is nonstandard or malformed
        sys_enc = None
    if sys_enc and sys_enc not in encodings:
        encodings.append(sys_enc)

    for enc in ["utf-8", "latin-1"]:
        if enc not in encodings:
            encodings.append(enc)

    return encodings


def split_locale(loc):
    """
    Split a locale specifier.  The general format is

    language[_territory][.codeset][@modifier] [charmap]

    For example:

    ca_ES.UTF-8@valencia UTF-8
    """

    def split(st, char):
        """
        Split a string `st` once by `char`; always return a two-element list
        even if the second element is empty.
        """
        split_st = st.split(char, 1)
        if len(split_st) == 1:
            split_st.append("")
        return split_st

    comps = {}
    work_st, comps["charmap"] = split(loc, " ")
    work_st, comps["modifier"] = split(work_st, "@")
    work_st, comps["codeset"] = split(work_st, ".")
    comps["language"], comps["territory"] = split(work_st, "_")
    return comps


def join_locale(comps):
    """
    Join a locale specifier split in the format returned by split_locale.
    """
    loc = comps["language"]
    if comps.get("territory"):
        loc += "_" + comps["territory"]
    if comps.get("codeset"):
        loc += "." + comps["codeset"]
    if comps.get("modifier"):
        loc += "@" + comps["modifier"]
    if comps.get("charmap"):
        loc += " " + comps["charmap"]
    return loc


def normalize_locale(loc):
    """
    Format a locale specifier according to the format returned by `locale -a`.
    """
    comps = split_locale(loc)
    comps["territory"] = comps["territory"].upper()
    comps["codeset"] = comps["codeset"].lower().replace("-", "")
    comps["charmap"] = ""
    return join_locale(comps)



import salt.config
from tests.support.mock import MagicMock, patch


def test___cli_path_is_expanded():
    defaults = salt.config.DEFAULT_MINION_OPTS.copy()
    overrides = {}
    with patch(
        "salt.utils.path.expand", MagicMock(return_value="/path/to/testcli")
    ) as expand_mock:
        opts = salt.config.apply_minion_config(overrides, defaults)
        assert expand_mock.called
        assert opts["__cli"] == "testcli"



"""
Tests for salt-run with show_jid
"""

import logging
import re

import pytest

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def salt_master(salt_factories):
    """
    Salt master with `show_jid: True`
    """
    config_defaults = {
        "show_jid": True,
    }
    salt_master = salt_factories.salt_master_daemon(
        "salt-run-show-jid-master", defaults=config_defaults
    )
    with salt_master.started():
        yield salt_master


@pytest.fixture(scope="module")
def salt_run_cli(salt_master):
    """
    The ``salt-run`` CLI as a fixture against the running master
    """
    assert salt_master.is_running()
    return salt_master.salt_run_cli(timeout=30)


def test_salt_run_show_jid(salt_run_cli):
    """
    Test that jid is output
    """
    ret = salt_run_cli.run("test.stdout_print")
    assert re.match(r"jid: \d+", ret.stdout)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.states.gnomedesktop
"""

import pytest

import salt.states.gnomedesktop as gnomedesktop


@pytest.fixture
def configure_loader_modules():
    return {gnomedesktop: {}}


def test_wm_preferences():
    """
    Test to sets values in the org.gnome.desktop.wm.preferences schema
    """
    name = "salt"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    assert gnomedesktop.wm_preferences(name) == ret


def test_desktop_lockdown():
    """
    Test to sets values in the org.gnome.desktop.lockdown schema
    """
    name = "salt"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    assert gnomedesktop.desktop_lockdown(name) == ret


def test_desktop_interface():
    """
    Test to sets values in the org.gnome.desktop.interface schema
    """
    name = "salt"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    assert gnomedesktop.desktop_interface(name) == ret



"""
Take data from salt and "return" it into a carbon receiver

Add the following configuration to the minion configuration file:

.. code-block:: yaml

    carbon.host: <server ip address>
    carbon.port: 2003

Errors when trying to convert data to numbers may be ignored by setting
``carbon.skip_on_error`` to `True`:

.. code-block:: yaml

    carbon.skip_on_error: True

By default, data will be sent to carbon using the plaintext protocol. To use
the pickle protocol, set ``carbon.mode`` to ``pickle``:

.. code-block:: yaml

    carbon.mode: pickle

You can also specify the pattern used for the metric base path (except for virt modules metrics):
    carbon.metric_base_pattern: carbon.[minion_id].[module].[function]

These tokens can used :
    [module]: salt module
    [function]: salt function
    [minion_id]: minion id

Default is :
    carbon.metric_base_pattern: [module].[function].[minion_id]

Carbon settings may also be configured as:

.. code-block:: yaml

    carbon:
      host: <server IP or hostname>
      port: <carbon port>
      skip_on_error: True
      mode: (pickle|text)
      metric_base_pattern: <pattern> | [module].[function].[minion_id]

Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location:

.. code-block:: yaml

    alternative.carbon:
      host: <server IP or hostname>
      port: <carbon port>
      skip_on_error: True
      mode: (pickle|text)

To use the carbon returner, append '--return carbon' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return carbon

To use the alternative configuration, append '--return_config alternative' to the salt command.

.. versionadded:: 2015.5.0

.. code-block:: bash

    salt '*' test.ping --return carbon --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return carbon --return_kwargs '{"skip_on_error": False}'

"""

import logging
import pickle
import socket
import struct
import time
from collections.abc import Mapping
from contextlib import contextmanager

import salt.returners
import salt.utils.jid

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "carbon"


def __virtual__():
    return __virtualname__


def _get_options(ret):
    """
    Returns options used for the carbon returner.
    """
    attrs = {"host": "host", "port": "port", "skip": "skip_on_error", "mode": "mode"}

    _options = salt.returners.get_returner_options(
        __virtualname__, ret, attrs, __salt__=__salt__, __opts__=__opts__
    )
    return _options


@contextmanager
def _carbon(host, port):
    """
    Context manager to ensure the clean creation and destruction of a socket.

    host
        The IP or hostname of the carbon server
    port
        The port that carbon is listening on
    """
    carbon_sock = None

    try:
        carbon_sock = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP
        )

        carbon_sock.connect((host, port))
    except OSError as err:
        log.error("Error connecting to %s:%s, %s", host, port, err)
        raise
    else:
        log.debug("Connected to carbon")
        yield carbon_sock
    finally:
        if carbon_sock is not None:
            # Shut down and close socket
            log.debug("Destroying carbon socket")

            carbon_sock.shutdown(socket.SHUT_RDWR)
            carbon_sock.close()


def _send_picklemetrics(metrics):
    """
    Format metrics for the carbon pickle protocol
    """

    metrics = [
        (metric_name, (timestamp, value)) for (metric_name, value, timestamp) in metrics
    ]

    data = pickle.dumps(metrics, -1)
    payload = struct.pack(b"!L", len(data)) + data

    return payload


def _send_textmetrics(metrics):
    """
    Format metrics for the carbon plaintext protocol
    """

    data = [" ".join(map(str, metric)) for metric in metrics] + [""]

    return "\n".join(data)


def _walk(path, value, metrics, timestamp, skip):
    """
    Recursively include metrics from *value*.

    path
        The dot-separated path of the metric.
    value
        A dictionary or value from a dictionary. If a dictionary, ``_walk``
        will be called again with the each key/value pair as a new set of
        metrics.
    metrics
        The list of metrics that will be sent to carbon, formatted as::

            (path, value, timestamp)
    skip
        Whether or not to skip metrics when there's an error casting the value
        to a float. Defaults to `False`.
    """
    log.trace(
        "Carbon return walking path: %s, value: %s, metrics: %s, timestamp: %s",
        path,
        value,
        metrics,
        timestamp,
    )
    if isinstance(value, Mapping):
        for key, val in value.items():
            _walk(f"{path}.{key}", val, metrics, timestamp, skip)
    elif isinstance(value, list):
        for item in value:
            _walk(f"{path}.{item}", item, metrics, timestamp, skip)

    else:
        try:
            val = float(value)
            metrics.append((path, val, timestamp))
        except (TypeError, ValueError):
            msg = (
                "Error in carbon returner, when trying to convert metric: "
                "{}, with val: {}".format(path, value)
            )
            if skip:
                log.debug(msg)
            else:
                log.info(msg)
                raise


def _send(saltdata, metric_base, opts):
    """
    Send the data to carbon
    """

    host = opts.get("host")
    port = opts.get("port")
    skip = opts.get("skip")
    metric_base_pattern = opts.get("carbon.metric_base_pattern")
    mode = opts.get("mode").lower() if "mode" in opts else "text"

    log.debug("Carbon minion configured with host: %s:%s", host, port)
    log.debug("Using carbon protocol: %s", mode)

    if not (host and port):
        log.error("Host or port not defined")
        return

    # TODO: possible to use time return from salt job to be slightly more precise?
    # convert the jid to unix timestamp?
    # {'fun': 'test.version', 'jid': '20130113193949451054', 'return': '0.11.0', 'id': 'salt'}
    timestamp = int(time.time())

    handler = _send_picklemetrics if mode == "pickle" else _send_textmetrics
    metrics = []
    log.trace("Carbon returning walking data: %s", saltdata)
    _walk(metric_base, saltdata, metrics, timestamp, skip)
    data = handler(metrics)
    log.trace("Carbon inserting data: %s", data)

    with _carbon(host, port) as sock:
        total_sent_bytes = 0
        while total_sent_bytes < len(data):
            sent_bytes = sock.send(data[total_sent_bytes:])
            if sent_bytes == 0:
                log.error("Bytes sent 0, Connection reset?")
                return

            log.debug("Sent %s bytes to carbon", sent_bytes)
            total_sent_bytes += sent_bytes


def event_return(events):
    """
    Return event data to remote carbon server

    Provide a list of events to be stored in carbon
    """
    opts = _get_options({})  # Pass in empty ret, since this is a list of events
    opts["skip"] = True
    for event in events:
        log.trace("Carbon returner received event: %s", event)
        metric_base = event["tag"]
        saltdata = event["data"].get("data")
        _send(saltdata, metric_base, opts)


def returner(ret):
    """
    Return data to a remote carbon server using the text metric protocol

    Each metric will look like::

        [module].[function].[minion_id].[metric path [...]].[metric name]

    """
    opts = _get_options(ret)
    metric_base = ret["fun"]
    # Strip the hostname from the carbon base if we are returning from virt
    # module since then we will get stable metric bases even if the VM is
    # migrate from host to host
    if not metric_base.startswith("virt."):
        metric_base += "." + ret["id"].replace(".", "_")

    saltdata = ret["return"]
    _send(saltdata, metric_base, opts)


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)



"""
.. versionadded:: 2015.5.0

Salt-ssh wrapper functions for the publish module.

Publish will never actually execute on the minions, so we just create new
salt-ssh calls and return the data from them.

No access control is needed because calls cannot originate from the minions.

.. versionchanged:: 3007.0

    In addition to SSH minions, this module can now also target regular ones.
"""

import copy
import logging
import time

import salt.client.ssh
import salt.daemons.masterapi
import salt.runner
import salt.utils.args
import salt.utils.json

log = logging.getLogger(__name__)


def _parse_args(arg):
    """
    yamlify `arg` and ensure its outermost datatype is a list
    """
    yaml_args = salt.utils.args.yamlify_arg(arg)

    if yaml_args is None:
        return []
    elif not isinstance(yaml_args, list):
        return [yaml_args]
    else:
        return yaml_args


def _publish(
    tgt,
    fun,
    arg=None,
    tgt_type="glob",
    returner="",
    timeout=None,
    form="clean",
    roster=None,
):
    """
    Publish a command "from the minion out to other minions". In reality, the
    minion does not execute this function, it is executed by the master. Thus,
    no access control is enabled, as minions cannot initiate publishes
    themselves.

    Salt-ssh publishes will default to whichever roster was used for the
    initiating salt-ssh call, and can be overridden using the ``roster``
    argument

    Returners are not currently supported

    The arguments sent to the minion publish function are separated with
    commas. This means that for a minion executing a command with multiple
    args it will look like this::

        salt-ssh system.example.com publish.publish '*' user.add 'foo,1020,1020'

    CLI Example:

    .. code-block:: bash

        salt-ssh system.example.com publish.publish '*' cmd.run 'ls -la /tmp'
    """
    if fun.startswith("publish."):
        log.info("Cannot publish publish calls. Returning {}")
        return {}

    # TODO: implement returners? Do they make sense for salt-ssh calls?
    if returner:
        log.warning("Returners currently not supported in salt-ssh publish")

    # Make sure args have been processed
    if arg is None:
        arg = []
    elif not isinstance(arg, list):
        # yamlify_arg does not operate on non-strings, which we need to JSON-encode
        arg = [salt.utils.json.dumps(salt.utils.args.yamlify_arg(arg))]
    else:
        arg = [
            salt.utils.json.dumps(y)
            for y in (salt.utils.args.yamlify_arg(x) for x in arg)
        ]
    if len(arg) == 1 and arg[0] is None:
        arg = []

    # Set up opts for the SSH object
    opts = copy.deepcopy(__context__["master_opts"])
    minopts = copy.deepcopy(__opts__)
    opts.update(minopts)
    if roster:
        opts["roster"] = roster
    if timeout:
        opts["timeout"] = timeout
    opts["argv"] = [fun] + arg
    opts["selected_target_option"] = tgt_type
    opts["tgt"] = tgt
    opts["arg"] = arg

    # Create the SSH object to handle the actual call
    ssh = salt.client.ssh.SSH(opts)

    # Run salt-ssh to get the minion returns
    rets = {}
    for ret in ssh.run_iter():
        rets.update(ret)

    if form == "clean":
        cret = {}
        for host in rets:
            if "return" in rets[host]:
                cret[host] = rets[host]["return"]
            else:
                cret[host] = rets[host]
        return cret
    for host in rets:
        if "return" in rets[host]:
            # The regular publish return just contains `ret`,
            # at least make it accessible like this as well
            rets[host]["ret"] = rets[host]["return"]
    return rets


def _publish_regular(
    tgt,
    fun,
    arg=None,
    tgt_type="glob",
    returner="",
    timeout=5,
    form="clean",
    wait=False,
):
    if fun.startswith("publish."):
        log.info("Cannot publish publish calls. Returning {}")
        return {}

    arg = _parse_args(arg)
    masterapi = salt.daemons.masterapi.RemoteFuncs(__context__["master_opts"])

    log.info("Publishing '%s'", fun)
    load = {
        "cmd": "minion_pub",
        "fun": fun,
        "arg": arg,
        "tgt": tgt,
        "tgt_type": tgt_type,
        "ret": returner,
        "tmo": timeout,
        "form": form,
        "id": __opts__["id"],
        "no_parse": __opts__.get("no_parse", []),
    }
    peer_data = masterapi.minion_pub(load)
    if not peer_data:
        return {}
    # CLI args are passed as strings, re-cast to keep time.sleep happy
    if wait:
        loop_interval = 0.3
        matched_minions = set(peer_data["minions"])
        returned_minions = set()
        loop_counter = 0
        while returned_minions ^ matched_minions:
            load = {
                "cmd": "pub_ret",
                "id": __opts__["id"],
                "jid": peer_data["jid"],
            }
            ret = masterapi.pub_ret(load)
            returned_minions = set(ret.keys())

            end_loop = False
            if returned_minions >= matched_minions:
                end_loop = True
            elif (loop_interval * loop_counter) > timeout:
                if not returned_minions:
                    return {}
                end_loop = True

            if end_loop:
                if form == "clean":
                    cret = {}
                    for host in ret:
                        cret[host] = ret[host]["ret"]
                    return cret
                else:
                    return ret
            loop_counter = loop_counter + 1
            time.sleep(loop_interval)
    else:
        time.sleep(float(timeout))
        load = {
            "cmd": "pub_ret",
            "id": __opts__["id"],
            "jid": peer_data["jid"],
        }
        ret = masterapi.pub_ret(load)
        if form == "clean":
            cret = {}
            for host in ret:
                cret[host] = ret[host]["ret"]
            return cret
        else:
            return ret
    return ret


def publish(
    tgt,
    fun,
    arg=None,
    tgt_type="glob",
    returner="",
    timeout=5,
    roster=None,
    ssh_minions=True,
    regular_minions=False,
):
    """
    Publish a command from the minion out to other minions. In reality, the
    minion does not execute this function, it is executed by the master. Thus,
    no access control is enabled, as minions cannot initiate publishes
    themselves.

    Salt-ssh publishes will default to whichever roster was used for the
    initiating salt-ssh call, and can be overridden using the ``roster``
    argument.

    Returners are not currently supported

    The tgt_type argument is used to pass a target other than a glob into
    the execution, the available options for SSH minions are:

    - glob
    - pcre
    - nodegroup
    - range

    Regular minions support all usual ones.

    .. versionchanged:: 2017.7.0
        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    The arguments sent to the minion publish function are separated with
    commas. This means that for a minion executing a command with multiple
    args it will look like this:

    .. code-block:: bash

        salt-ssh system.example.com publish.publish '*' user.add 'foo,1020,1020'
        salt-ssh system.example.com publish.publish '127.0.0.1' network.interfaces '' roster=scan

    CLI Example:

    .. code-block:: bash

        salt-ssh system.example.com publish.publish '*' cmd.run 'ls -la /tmp'


    .. admonition:: Attention

        If you need to pass a value to a function argument and that value
        contains an equal sign, you **must** include the argument name.
        For example:

        .. code-block:: bash

            salt-ssh '*' publish.publish test.kwarg arg='cheese=spam'

        Multiple keyword arguments should be passed as a list.

        .. code-block:: bash

            salt-ssh '*' publish.publish test.kwarg arg="['cheese=spam','spam=cheese']"


    tgt
        The target specification.

    fun
        The execution module to run.

    arg
        A list of arguments to pass to the module.

    tgt_type
        The matcher to use. Defaults to ``glob``.

    returner
        A returner to use.

    timeout
        Timeout in seconds. Defaults to 5.

    roster
        Override the roster for SSH minion targets. Defaults to the one
        used for initiating the salt-ssh call.

    ssh_minions
        .. versionadded:: 3007.0
        Include SSH minions in the possible targets. Defaults to true.

    regular_minions
        .. versionadded:: 3007.0
        Include regular minions in the possible targets. Defaults to false.
    """
    rets = {}
    if regular_minions:
        rets.update(
            _publish_regular(
                tgt,
                fun,
                arg=arg,
                tgt_type=tgt_type,
                returner=returner,
                timeout=timeout,
                form="clean",
                wait=True,
            )
        )
    if ssh_minions:
        rets.update(
            _publish(
                tgt,
                fun,
                arg=arg,
                tgt_type=tgt_type,
                returner=returner,
                timeout=timeout,
                form="clean",
                roster=roster,
            )
        )
    return rets


def full_data(
    tgt,
    fun,
    arg=None,
    tgt_type="glob",
    returner="",
    timeout=5,
    roster=None,
    ssh_minions=True,
    regular_minions=False,
):
    """
    Return the full data about the publication, this is invoked in the same
    way as the publish function

    CLI Example:

    .. code-block:: bash

        salt-ssh system.example.com publish.full_data '*' cmd.run 'ls -la /tmp'

    .. admonition:: Attention

        If you need to pass a value to a function argument and that value
        contains an equal sign, you **must** include the argument name.
        For example:

        .. code-block:: bash

            salt-ssh '*' publish.full_data test.kwarg arg='cheese=spam'

    """
    rets = {}
    if regular_minions:
        rets.update(
            _publish_regular(
                tgt,
                fun,
                arg=arg,
                tgt_type=tgt_type,
                returner=returner,
                timeout=timeout,
                form="full",
                wait=True,
            )
        )
    if ssh_minions:
        rets.update(
            _publish(
                tgt,
                fun,
                arg=arg,
                tgt_type=tgt_type,
                returner=returner,
                timeout=timeout,
                form="full",
                roster=roster,
            )
        )
    return rets


def runner(fun, arg=None, timeout=5):
    """
    Execute a runner on the master and return the data from the runner function

    CLI Example:

    .. code-block:: bash

        salt-ssh '*' publish.runner jobs.lookup_jid 20140916125524463507
    """
    # Form args as list
    if not isinstance(arg, list):
        arg = [salt.utils.args.yamlify_arg(arg)]
    else:
        arg = [salt.utils.args.yamlify_arg(x) for x in arg]
    if len(arg) == 1 and arg[0] is None:
        arg = []

    # Create and run the runner
    runner = salt.runner.RunnerClient(__opts__["__master_opts__"])
    return runner.cmd(fun, arg)
